#!/usr/bin/env python

# A Gimp plugin to replicate a path by applying one affine transformation
# repeatedly when given is one such transformed path.
# So, the transformation is not input. Instead input is the path and
# one transformed copy.
#
# History:
# v0.1: 2021-12-21: Plugin "Replicate path by affine mapping"
# v0.2: 2021-12-21: New wording in the GUI.

# (c) Markku Koppinen 2021
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published
#   by the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This very file is the complete source code to the program.
#
#   If you make and redistribute changes to this code, please mark it
#   in reasonable ways as different from the original version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   The GPL v3 licence is available at: https://www.gnu.org/licenses/gpl-3..en.html'

from __future__ import division, print_function
from gimpfu import *
from math import *
from cmath import exp as cexp
from copy import deepcopy

#==============================================================
#                         Draw
#==============================================================
# ------------------------
# gimp_draw_vectors_object
# ------------------------
# The Gimp version of the routine 'drawing_routine_for_bezier':
# Insert into Gimp the input vectors object, and make it visible if
# visible=True.
# Args:
# - image: Gimp's image (obtained usually from GUI);
# - vectors_object: a vectors object in Gimp;
# - visible: Boolean.
# Returns the vectors object in Gimp, and makes it visible if visible=True.
def gimp_draw_vectors_object(image, vectors_object, visible=True):
    # Insert a Gimp vectors object in Gimp.
    # Adapted from http://registry.gimp.org/files/lqr_wpset-0.14.py.txt
    # by Mike Kazantsev.
    def gimp_insert_vectors(image, vectors_object, position=0):
        if gimp.version >= (2, 8, 0):
            # 2.7.0 has gimp-image-insert-vectors, but fails with
            # parent=None
            parent = None
            return pdb.gimp_image_insert_vectors(
                     image, vectors_object, parent, position)
        else:
            return pdb.gimp_image_add_vectors(
                     image, vectors_object, position)# Deprecated!
    gimp_insert_vectors(image, vectors_object)
    vectors_object.visible = visible
    return vectors_object

#==============================================================
#                     Plane vectors
#==============================================================

def collinear(points): # points: [complex] (The plane R2 is the complex plane.)
    ZERO = 1e-14
    n = len(points)
    if n < 3:
        return True
    for i in range(n):
        for j in range(i+1,n):
            for k in range(j+1,n):
                a,b,c = points[i], points[j], points[k]
                ab = -a+b
                ac = -a+c
                bc = -b+c
                if abs(cross(ab,ac)) > ZERO:
                    return False
                if abs(cross(ab,bc)) > ZERO:
                    return False
                if abs(cross(ac,bc)) > ZERO:
                    return False
    return True

# ---------------
# make_similitude
# ---------------
# Make similitude where z0,z1 are mapped onto w0,w1.
# Args:
# - z0,z1,w0,w1: complex
# Returns:
# - callable (complex -> complex)
def make_similitude(z0,z1,w0,w1):
    def similitude(z):
        return (w1*(z-z0) - w0*(z-z1)) / (-z0+z1)
    return similitude

# ---------------
# make_affine_map
# ---------------
# Affine map sending three given points u0,u1,u2 to another
# three given points v0,v1,v2. 
# Args:
# - base:   [complex,complex,complex]
# - target: [complex,complex,complex]
# Returns:
# - callable (complex -> complex)
def make_affine_map(base, target):
    u0,u1,u2 = base
    v0,v1,v2 = target
    if collinear([u0,u1,u2]):
        raise Exception("make_affine_map: Anchors of Base are collinear.")
    u10 = u1-u0
    u20 = u2-u0
    v10 = v1-v0
    v20 = v2-v0
    A = Matrix2x2(u10.real, u10.imag, 
                  u20.real, u20.imag)
    B = Matrix2x2(v10.real, v10.imag, 
                  v20.real, v20.imag)
    try:
        AB = A.inverse() * B
    except Exception as e:
        raise Exception("make_affine_map: "+str(e))
    def pm(x):
        return v0 + rv_times_M2x2((x-u0),AB)
    #def jac(x):
    #    return AB.transpose()
    return pm # Point map only

# -------------
# rv_times_M2x2
# -------------
# RowVector times Matrix2x2.
# Args:
# - rv:     complex
# - matrix: Matrix2x2
# Returns:
# - complex
def rv_times_M2x2(rv,matrix):
    return complex(rv.real*matrix[0,0] + rv.imag*matrix[1,0],
                     rv.real*matrix[0,1] + rv.imag*matrix[1,1])

class Matrix2x2(object):
    """2x2-matrix
    Initialize m = Matrix2x2(1,2,3,4)
    Get elements as m[i,j]
    Set elements as m[i,j]=value
    Note: Works with real or complex matrices.
    """
    def __init__(self,a00,a01,a10,a11):
        #self._elements = [float(a00),float(a01),float(a10),float(a11)]
        self._elements = [a00,a01,a10,a11]
    def __getitem__(self,indexes): # indexes = [i,j]
        i,j = indexes
        if not (0<=i<=1 and 0<=j<=1):
            raise Exception("Matrix2x2 get: Index out of range")
        return self._elements[2*i+j]
    def __setitem__(self, indexes, value):
        i,j = indexes
        if not (0<=i<=1 and 0<=j<=1):
            raise Exception("Matrix2x2 set: Index out of range")
        self._elements[2*i+j] = value
    def transpose(self):
        a,b,c,d = self._elements
        return Matrix2x2(a,c,b,d)
    def scale(self, coeff):
        a,b,c,d = [x*coeff for x in self._elements]
        return Matrix2x2(a,b,c,d)
    def inverse(self):
        a,b,c,d = self._elements
        try:
            idet = 1./(a*d-b*c)
        except ZeroDivisionError:
            raise Exception("Matrix2x2 inverse: Zero determinant")
        A,B,C,D = [x*idet for x in self._elements]
        return Matrix2x2(D,-B,-C,A)
    def __add__(self,other):
        a,b,c,d = self._elements
        A,B,C,D = other._elements
        return Matrix2x2(a+A,b+B,c+C,d+D)
    def __mul__(self,other):
        return Matrix2x2(self[0,0]*other[0,0] + self[0,1]*other[1,0],
                         self[0,0]*other[0,1] + self[0,1]*other[1,1],
                         self[1,0]*other[0,0] + self[1,1]*other[1,0],
                         self[1,0]*other[0,1] + self[1,1]*other[1,1])
    def __str__(self):
        def str_float(x): return '{:1.3f}'.format(x)
        a00 = str_float(self[0,0])
        a01 = str_float(self[0,1])
        a10 = str_float(self[1,0])
        a11 = str_float(self[1,1])
        return '['+a00+','+a01+']' +'\n['+a10+','+a11+']\n'

# v dot (w perpendicular)
def cross(v,w):
    return (v*w.conjugate()).imag

# -----------------
# largest_triangles
# -----------------
# Given a list of ControlPoints in the plane,
# find among them the triangles with largest absolute areas.
# If found non-zero triangle, return a list of the largest triangles
# (ControlPoint triples) that deviate at most by 'tolerance' from the very
# largest triangle.
# If nothing non-zero is found, None is returned; this means that the points
# are collinear.
# The returned list is sorted from largest down.
# Args:
# - points:    [ControlPoint]
# - tolerance: float
# Returns: None or
# - [(ControlPoint,ControlPoint,ControlPoint), ...] (list of point triples)
def largest_triangles(points, tolerance=0.1):
    ZERO = 1e-6
    points = list(set(points)) # Remove duplicates
    point3_areas = [] # [((complex,complex,complex),float)] = [(triangle, area)]
    for i in range(len(points)):
        x = points[i]
        for j in range(i+1,len(points)):
            y = points[j]
            for k in range(j+1,len(points)):
                z = points[k]
                area = abs(cross(y.point-x.point, z.point-x.point))/2
                point3_areas.append( ((x,y,z),area) )
    point3_areas.sort(key=(lambda pa:-pa[1]))
    largest_area = point3_areas[0][1]
    if largest_area < ZERO:
        return None
    accept_limit = (1-tolerance) * largest_area
    accepted = [p3a for p3a in point3_areas if p3a[1] >= accept_limit]
    return [p3 for p3,area in accepted]

# -----------------
# most_distant_pair
# -----------------
# Given a list of ControlPoints in the plane,
# find among them the most widely separated pair.
# If found non-zero, return a list of the most widely separated pairs
# (ControlPoint doubles) that deviate at most by 'tolerance' from the very
# widest pair.
# If nothing non-zero is found, None is returned; this means that the points
# are coincident.
# The returned list is sorted from largest down.
# Args:
# - points:    [ControlPoint]
# - tolerance: float
# Returns: None or
# - [(ControlPoint,ControlPoint), ...] (list of point doubles)
def most_distant_pair(points, tolerance=0.1):
    ZERO = 1e-6
    points = list(set(points)) # Remove duplicates
    
    point2_distances = [] # [((complex,complex),float)] = [(point, area)]
    for i in range(len(points)):
        x = points[i]
        for j in range(i+1,len(points)):
            y = points[j]
            distance = abs(y.point-x.point)
            point2_distances.append( ((x,y),distance) )
    point2_distances.sort(key=(lambda pa:-pa[1]))
    largest_distance = point2_distances[0][1]
    if largest_distance < ZERO:
        return None
    accept_limit = (1-tolerance) * largest_distance
    accepted = [p2d for p2d in point2_distances if p2d[1] >= accept_limit]
    return [p2 for p2,distance in accepted]


#==============================================================
#                        Paths
#==============================================================

ANCHOR = 0
HANDLE = 1

OPEN = 0
CLOSED = 1

class ControlPoint(object):
    def __init__(self,
                 point,           # complex
                 anchor_or_handle # ANCHOR or HANDLE
                 ):
        self.point = point
        self.anchor_or_handle = anchor_or_handle
    def __str__(self):
        s = "ControlPoint"
        s += "\n  point: "+str(self.point)
        s += "\n  type:  "+("ANCHOR" if self.anchor_or_handle == ANCHOR else "HANDLE")
        return s

class Stroke(object):
    def __init__(self,
                 pointlist,     # [ControlPoint]
                 open_or_closed # OPEN or CLOSED
                 ):
        self.pointlist = pointlist
        self.open_or_closed = open_or_closed
    def __str__(self):
        s = "Stroke"
        s += "  pointlist: "
        for cp in self.pointlist: s += "\n    "+str(cp)
        s += "\n  type:  "+("OPEN" if self.open_or_closed == OPEN else "CLOSED")
        return s

# -----------------
# get_controlpoints
# -----------------
# From a path extract anchors and handles, one list for each stroke.
# The plane is viewed as the complex number plane, so all points are
# complex numbers.
# Args:
# - path: gimp.Vectors
# Returns:
# - [Stroke] (one item for each stroke)
def get_controlpoints(path):
    cps_list = []
    for stroke in path.strokes:
        cps = []
        xy_list = stroke.points[0]
        for i in range(0,len(xy_list),6):
            cps.append(ControlPoint(point=complex(xy_list[i], xy_list[i+1]),
                                       anchor_or_handle=HANDLE))
            cps.append(ControlPoint(point=complex(xy_list[i+2], xy_list[i+3]),
                                       anchor_or_handle=ANCHOR))
            cps.append(ControlPoint(point=complex(xy_list[i+4], xy_list[i+5]),
                                       anchor_or_handle=HANDLE))
        closed = (CLOSED if stroke.points[1] else OPEN)
        cps_list.append(Stroke(pointlist=cps, open_or_closed=closed))
    return cps_list

# ---------
# copy_path
# ---------
# Copy but do not insert in Gimp.
def copy_path(image, path):
    path_copied = pdb.gimp_vectors_new(image, path.name+'|copy')
    for stroke in path.strokes:
        xy_list, closed = stroke.points
        pdb.gimp_vectors_stroke_new_from_points(
                             path_copied, 0, len(xy_list), xy_list, closed)
    return path_copied

# -------------------
# path_affine_mapping
# -------------------
# Given a path and an affine mapping (complex -> complex),
# transform the path with that mapping.
# Args:
# - image
# - path:       gimp.Vectors
# - affine_map: callable (complex -> complex)
# Returns:
# - gimp.Vectors
def path_affine_mapping(image, path, affine_map):
    new_strokelist = []
    path_transformed = pdb.gimp_vectors_new(image, path.name+'|transformed')
    for stroke in path.strokes:
        cps = []
        xy_list, closed = stroke.points
        for i in range(0,len(xy_list),2):
            cps.append(complex(xy_list[i], complex(xy_list[i+1])))
        cps_transformed = [affine_map(cp) for cp in cps]
        xy_transformed = []
        for cp in cps_transformed:
            xy_transformed += [cp.real, cp.imag]
        pdb.gimp_vectors_stroke_new_from_points(
                path_transformed, 0, len(xy_transformed), xy_transformed, closed)
    return path_transformed

# -------------------------
# strokelist_affine_mapping
# -------------------------
# Given a stroke list and an affine mapping (complex -> complex),
# transform the stroke list with that mapping.
# The strokes in the list are in Gimp's form ([x,y,x,y,x,y,...], closed).
# Return in the same form.
# Args:
# - strokelist: [([x,y,x,y,x,y,...], closed)]
# - affine_map: callable (complex -> complex)
# Returns:
# - new strokelist: [([x,y,x,y,x,y,...], closed)]
def strokelist_affine_mapping(strokelist, affine_map):
    new_strokelist = []
    for xy_list, closed in strokelist:
        cps = []
        for i in range(0,len(xy_list),2):
            cps.append(complex(xy_list[i], complex(xy_list[i+1])))
        cps_transformed = [affine_map(cp) for cp in cps]
        xy_transformed = []
        for cp in cps_transformed:
            xy_transformed += [cp.real, cp.imag]
        new_strokelist.append((xy_transformed, closed))
    return new_strokelist

#==============================================================
#                        Get affine mapping
#==============================================================

# ------------------
# get_affine_mapping
# ------------------
# Given control points of two paths,
# the second obtained by some affine transformation from the first,
# find that affine transformation.
# Each of the two paths is represented as list of items of type Stroke,
# one for each stroke.
# Args:
# - stroke_list_ori: [Stroke] (original path)
# - stroke_list_tra: [Stroke] (transformed path)
# Returns:
# - callable (complex -> complex)
def get_affine_mapping(stroke_list_ori, stroke_list_tra):
    ACCEPT_ERROR = 1e-3
    TOLERANCE = 0.1
    # Put all items:ControlPoint in one list
    cps_all_ori = []
    for stroke in stroke_list_ori:
        cps_all_ori += stroke.pointlist
    cps_all_tra = []
    for stroke in stroke_list_tra:
        cps_all_tra += stroke.pointlist
    p3_list = largest_triangles(cps_all_ori, tolerance=TOLERANCE)
    #print("1: len(p3_list) = "+str(len(p3_list)))
    if not (p3_list is None): # Found proper triangle: not all on a straight line
        p3_list_transformed = largest_triangles(cps_all_tra, tolerance=TOLERANCE)
        #print("1: len(p3_list_transformed) = "+str(len(p3_list_transformed)))
        if p3_list_transformed is None:
            raise Exception(NO_AFFINE_MAP_MESSAGE)
        permutations = []
        for w0,w1,w2 in p3_list_transformed:
            permutations += [(w0,w1,w2),
                             (w1,w2,w0),
                             (w2,w0,w1),
                             (w0,w2,w1),
                             (w2,w1,w0),
                             (w1,w0,w2)]
        #print("1: len(permutations) = "+str(len(permutations)))
        # Construct list of candidate affine maps:
        affine_map_list = []
        for z0,z1,z2 in p3_list:
            for w0,w1,w2 in permutations:
                if ((z0.anchor_or_handle == w0.anchor_or_handle)
                        and (z1.anchor_or_handle == w1.anchor_or_handle)
                        and (z2.anchor_or_handle == w2.anchor_or_handle)):
                    map = make_affine_map([z0.point, z1.point, z2.point],
                                                  [w0.point, w1.point, w2.point])
                    affine_map_list.append(map)
    else: # All on a straight line
        p2_list = most_distant_pair(cps_all_ori, tolerance=TOLERANCE)
        #print("2: len(p2_list) = "+str(len(p2_list)))
        if not (p2_list is None): # Found proper pair: not all in a single point
            p2_list_transformed = most_distant_pair(cps_all_tra, tolerance=TOLERANCE)
            #print("2: len(p2_list_transformed) = "+str(len(p2_list_transformed)))
            if p2_list_transformed is None:
                raise Exception(NO_AFFINE_MAP_MESSAGE)
            permutations = []
            for w0,w1 in p2_list_transformed:
                permutations += [(w0,w1), (w1,w0)]
            #print("2: len(permutations) = "+str(len(permutations)))
            # Construct list of candidate affine maps (now these are similitudes):
            affine_map_list = []
            for z0,z1 in p2_list:
                for w0,w1 in permutations:
                    if ((z0.anchor_or_handle == w0.anchor_or_handle)
                            and (z1.anchor_or_handle == w1.anchor_or_handle)):
                        map = make_similitude(z0.point, z1.point, w0.point, w1.point)
                        affine_map_list.append(map)
        else: # All in a single point (anchors and handles)
            anchor_ori = cps_all_ori[1].point
            anchor_tra = cps_all_tra[1].point
            def map(z): # translation
                return z + anchor_tra - anchor_ori
            affine_map_list = [map]
    # From the list we have to choose the right one (though it is not unique).
    # This would be easy if we could trust that when the transformed path was
    # constructed from the original path by an affine map, certain simple
    # assumptions held (see the comments in 'simplistic_strokelist_error').
    # We make a try of this simplistic scenario first:
    maps_from_simple = []
    for map in affine_map_list:
        if ACCEPT_ERROR >= simplistic_strokelist_error(map,
                                                       stroke_list_ori,
                                                       stroke_list_tra,
                                                       ACCEPT_ERROR):
            maps_from_simple.append(map)
    if len(maps_from_simple) != 0: # Found good maps from the simplistic assumption:
        affine_map_list = maps_from_simple
    else: # The simplistic assumption did not work.
        # So we go now through the strokelists one stroke at a time.
        open_strokes_ori = [stroke for stroke in stroke_list_ori if stroke.open_or_closed == OPEN]
        closed_strokes_ori = [stroke for stroke in stroke_list_ori if stroke.open_or_closed == CLOSED]
        open_strokes_tra = [stroke for stroke in stroke_list_tra if stroke.open_or_closed == OPEN]
        closed_strokes_tra = [stroke for stroke in stroke_list_tra if stroke.open_or_closed == CLOSED]
        #
        # From the list affine_map_list exclude the maps which do not map open
        # original strokes close enough to some transformed stroke:
        if len(open_strokes_ori) != 0:
            map_errors = []
            for map in affine_map_list:
                map_error = 0
                for stroke_ori in open_strokes_ori:
                    stroke_error = 0
                    # Search for acceptable image from among the transformed strokes:
                    for stroke_tra in open_strokes_tra:
                        err = open_stroke_error(map, stroke_ori, stroke_tra, ACCEPT_ERROR)
                        if err <= ACCEPT_ERROR: # found
                            stroke_error = max(err, stroke_error)
                            found = True
                            break
                    else: # No break, not found acceptable => the map not acceptable!
                        found = False
                        break
                    map_error = max(map_error, stroke_error)
                if found: # Map acceptable: store
                    map_errors.append((map, map_error))
                else: # Map not acceptable: skip
                    continue
            affine_map_list = [map for map,err in map_errors]
        # From the list affine_map_list exclude the maps which do not map closed
        # original strokes close enough to some transformed stroke:
        if len(closed_strokes_ori) != 0:
            map_errors = []
            for map in affine_map_list:
                map_error = 0
                for stroke_ori in closed_strokes_ori:
                    stroke_error = 0
                    # Search for acceptable image from among the transformed strokes:
                    for stroke_tra in closed_strokes_tra:
                        err = closed_stroke_error(map, stroke_ori, stroke_tra, ACCEPT_ERROR)
                        if err <= ACCEPT_ERROR: # found
                            stroke_error = max(err, stroke_error)
                            found = True
                            break
                    else: # No break, not found acceptable => the map not acceptable!
                        found = False
                        break
                    map_error = max(map_error, stroke_error)
                if found: # Map acceptable: store
                    map_errors.append((map, map_error))
                else: # Map not acceptable: skip
                    continue
            affine_map_list = [map for map,err in map_errors]
    if len(affine_map_list) == 0:
        raise Exception(NO_AFFINE_MAP_MESSAGE)
    # Finally, from the surviving candidates, choose the one which means
    # the shortest movement (ad hoc choice):
    map_movement = [] # (map, error, matrix)
    for map in affine_map_list:
        movement = sum(abs(cp.point - map(cp.point)) for cp in cps_all_ori)
        map_movement.append((map, movement))
    affine_map, movement = min(map_movement, key=(lambda me:me[1]))
    return affine_map

# ---------------------------
# simplistic_strokelist_error
# ---------------------------
# Given a map and stroke lists strokelist_ori,strokelist_tra:[Stroke],
# measure some kind of error for the condition map(strokelist_ori) = strokelist_tra.
# Here we work with the simplistic assumption that strokelist_tra is supposed to
# come from strokelist_ori by an affine transformation in such a way that
# the following are preserved:
# - ordering of the strokes;
# - for open strokes the direction;
# - for closed strokes both the direction and the starting anchor.
# For large error just cut work short and return 1000.
# Args:
# - map:            callable (complex -> complex)
# - strokelist_ori: [Stroke]
# - strokelist_tra: [Stroke]
# - accept_error:   float
# Returns:
# - float
def simplistic_strokelist_error(map, strokelist_ori, strokelist_tra, accept_error):
    if len(strokelist_ori) != len(strokelist_tra):
        return 1000
    total_error = 0
    for stroke_index, stroke_ori in enumerate(strokelist_ori):
        stroke_tra = strokelist_tra[stroke_index]
        points_ori = stroke_ori.pointlist
        points_tra = stroke_tra.pointlist
        if len(points_ori) != len(points_tra):
            return 1000
        for point_index, cp_ori in enumerate(points_ori):
            map_point_ori = map(cp_ori.point)
            point_tra = points_tra[point_index].point
            err = abs(map_point_ori - point_tra)
            if err > accept_error:
                return 1000
            total_error = max(err, total_error)
    return total_error

# -----------------
# open_stroke_error
# -----------------
# Given a map and two open strokes stroke_ori,stroke_tra:Stroke,
# measure some kind of error for the condition map(stroke_ori) = stroke_tra.
# For large error just cut work short and return 1000.
# Args:
# - map:          callable (complex -> complex)
# - stroke_ori:   Stroke
# - stroke_tra:   Stroke
# - accept_error: float
# Returns:
# - float
def open_stroke_error(map, stroke_ori, stroke_tra, accept_error):
    if len(stroke_ori.pointlist) != len(stroke_tra.pointlist):
        return 1000
    total_error = 0
    tra = stroke_tra.pointlist
    for i,cp_ori in enumerate(stroke_ori.pointlist):
        z = map(cp_ori.point)
        cp_tra = tra[i]
        err = abs(z - cp_tra.point)
        if err > accept_error:
            break # Abandon this try
        total_error = max(err, total_error)
    else: # No break: accept
        return total_error
    # Try reversed image stroke:
    total_error = 0
    tra = stroke_tra.pointlist[::-1]
    for i,cp_ori in enumerate(stroke_ori.pointlist):
        z = map(cp_ori.point)
        cp_tra = tra[i]
        err = abs(z - cp_tra.point)
        if err > accept_error:
            return 1000
        total_error = max(err, total_error)
    return total_error

# -------------------
# closed_stroke_error
# -------------------
# Given a map and two closed strokes stroke_ori,stroke_tra:Stroke,
# measure some kind of error for the condition map(stroke_ori) = stroke_tra.
# For large error just cut work short and return 1000.
# Args:
# - map:          callable (complex -> complex)
# - stroke_ori:   Stroke
# - stroke_tra:   Stroke
# - accept_error: float
# Returns:
# - float
def closed_stroke_error(map, stroke_ori, stroke_tra, accept_error):
    if len(stroke_ori.pointlist) != len(stroke_tra.pointlist):
        return 1000
    total_error = 0
    tra = stroke_tra.pointlist
    # For closed strokes we must take possible rotations into account:
    for rotate in range(0, len(stroke_tra.pointlist), 3):
        for i,cp_ori in enumerate(stroke_ori.pointlist):
            z = map(cp_ori.point)
            cp_tra = tra[(i + rotate) % len(stroke_tra.pointlist)]
            err = abs(z - cp_tra.point)
            if err > accept_error:
                break # Abandon this try
            total_error = max(err, total_error)
        else: # No break: accept
            return total_error
    # Try reversed image stroke:
    total_error = 0
    tra = stroke_tra.pointlist[::-1]
    for rotate in range(0, len(stroke_tra.pointlist), 3):
        for i,cp_ori in enumerate(stroke_ori.pointlist):
            z = map(cp_ori.point)
            cp_tra = tra[(i + rotate) % len(stroke_tra.pointlist)]
            err = abs(z - cp_tra.point)
            if err > accept_error:
                return 1000
            total_error = max(err, total_error)
    return total_error

#==============================================================
#                        Main
#==============================================================

include_options = [ # (description, identifier)
        ('Include both',          0),
        #('Include only the seed', 1),
        ('Include only the auxiliary path', 1),
        ('Include neither',       2),
        ]

NO_AFFINE_MAP_MESSAGE = ("No affine transformation was found."
                         +"\nPerhaps you chose the wrong transformed path"
                         +"\nor used in its construction some other tool than"
                         +"\nMove, Scale, Rotate, Share, or Flip?")

def replicate_by_affine_mapping_main(image,
                                  path,
                                  transformed_path,
                                  ncopies,
                                  merge,
                                  include_option
                                  ):
    include = include_options[include_option][1]
    cps_list_original = get_controlpoints(path)
    cps_list_transformed = get_controlpoints(transformed_path)
    affine_map = get_affine_mapping(cps_list_original, cps_list_transformed)
    pdb.gimp_image_undo_group_start(image)
    if merge:
        new_path = pdb.gimp_vectors_new(image, path.name+'|transformed')
        for stroke in path.strokes:
            xy_list, closed = stroke.points
            if include == 0:
                pdb.gimp_vectors_stroke_new_from_points(
                    new_path, 0, len(xy_list), xy_list, closed)
        prev_strokelist = [] # [([x,y,x,y,...], closed)]
        for stroke in transformed_path.strokes:
            prev_strokelist.append(stroke.points)
            xy_list, closed = stroke.points
            if include <= 1:
                pdb.gimp_vectors_stroke_new_from_points(
                    new_path, 0, len(xy_list), xy_list, closed)
        for i in range(ncopies-2+include):
            new_strokelist = strokelist_affine_mapping(prev_strokelist, affine_map)
            for xy_list, closed in new_strokelist:
                pdb.gimp_vectors_stroke_new_from_points(
                    new_path, 0, len(xy_list), xy_list, closed)
            prev_strokelist = new_strokelist
        new_path.name = path.name+"|replicate"
        gimp_draw_vectors_object(image, new_path, visible=True)
    else: # Do not merge
        if include == 0:
            new_path = copy_path(image, path)
        elif include == 1:
            new_path = copy_path(image, transformed_path)
        else:
            new_path = path_affine_mapping(image, transformed_path, affine_map)
        for i in range(ncopies):
            new_path.name = path.name+"|replica "+str(i)
            last_drawn = gimp_draw_vectors_object(image, new_path, visible=True)
            new_path = path_affine_mapping(image, new_path, affine_map)
        new_path = last_drawn
    pdb.gimp_image_undo_group_end(image)
    return new_path

#======================================================
#                    Registration
#======================================================

versionnumber = "0.2"
procedure_author = "Markku Koppinen"
procedure_copyright = procedure_author
procedure_date = "2021"
image_types = "*"
menupath = '<Vectors>/Tools/Replicate'


procedure_name  = "replicate_path_by_affine_mapping"
procedure_blurb = ("Replicate a path by applying an affine mapping repeatedly,"
                   +"\ninputting one such transformed copy as an auxiliary path."
                   +"\n(Version "+versionnumber+")"
                   )
procedure_help  = ("Replicate a path by applying an affine mapping repeatedly."
                   +"\nThe affine mapping is not input; rather, the path and "
                   +"one transformed copy of it are input, and the plugin "
                   +"recognizes the transformation from them (correctly if it is unique)."
                   +"\nAllowed transformations are the affine:"
                   +"\nany combination of the tools"
                   +"\nMove, Scale, Rotate, Shear, Flip."
                   )

procedure_label = "Replicate path by affine mapping"

procedure_function = replicate_by_affine_mapping_main

register(
    procedure_name,
    procedure_blurb,
    procedure_help,
    procedure_author,
    procedure_copyright,
    procedure_date,
    procedure_label,
    image_types,
    [
      (PF_IMAGE, "image", "Input image", None),
      #(PF_VECTORS, "path", ("The primary input path:"
      #                      +"\nthe path to be replicated"
      #                      ), None),
      (PF_VECTORS, "path", ("The path to be replicated:"
                            #+"\nthe path to be replicated"
                            ), None),
      (PF_VECTORS, "transformed_path", 
                         ("The auxiliary path:"
                         +"\na path transformed from the main input path."
                         +"\n(Allowed transformations: Move, Scale, Rotate, Shear, Flip"
                         +"\nand any combination)"
                         #+"\nMove, Scale, Rotate, Shear, Flip)"
                         ), None),
      (PF_INT,   "ncopies", "Number of copies ", 10),
      (PF_BOOL,  "merge", "Merge down?", True),
      #(PF_OPTION, 'include_option','Include the original path and the seed in the output?',
      (PF_OPTION, 'include_option','Include the main input path and the auxiliary path in the output?',
                   0,
                  [case[0] for case in include_options]), # Descriptions of cases
    ],
    [
        (PF_VECTORS, "new_path", "New path"),
    ],
    procedure_function,
    menu=menupath)

main()

